unit PMX.Model;

interface

uses
  System.SysUtils, System.Classes;

type
  PmxByte = UInt8;
  PmxSByte = Int8;
  PmxUShort = UInt16;
  PmxShort = Int16;
  PmxUInt = UInt32;
  PmxInt = Int32;
  PmxFloat = Single;
  PmxFloat2 = packed record
    X, Y: PmxFloat;
  end;
  PmxFloat3 = packed record
    X, Y, Z: PmxFloat;
  end;
  PmxFloat4 = packed record
    X, Y, Z, W: PmxFloat;
  end;
  TPmxInt3 = packed record
    X, Y, Z: PmxInt;
  end;
  PPmxInt3 = ^TPmxInt3;

  TPmxBoneWeightMethod = (wmBDEF1, wmBDEF2, wmBDEF4, wmSDEF, wmQDEF);

  TPmxBoneWeightBDEF1 = record
    BoneIndex: PmxInt;
  end;

  TPmxBoneWeightBDEF2 = record
    Bone1Index: PmxInt;
    Bone2Index: PmxInt;
    Bone1Weight: PmxFloat;
  end;

  TPmxBoneWeightBDEF4 = record
    Bone1Index: PmxInt;
    Bone2Index: PmxInt;
    Bone3Index: PmxInt;
    Bone4Index: PmxInt;
    Bone1Weight: PmxFloat;
    Bone2Weight: PmxFloat;
    Bone3Weight: PmxFloat;
    Bone4Weight: PmxFloat;
  end;

  TPmxBoneWeightSDEF = record
    Bone1Index: PmxInt;
    Bone2Index: PmxInt;
    Bone1Weight: PmxFloat;
    SDEF_C: PmxFloat3;
    SDEF_R0: PmxFloat3;
    SDEF_R1: PmxFloat3;
  end;

  TPmxBoneWeightQDEF = TPmxBoneWeightBDEF4;

  TPmxBoneWeight = record
    case Integer of
      0: (BDEF1: TPmxBoneWeightBDEF1);
      1: (BDEF2: TPmxBoneWeightBDEF2);
      2: (BDEF4: TPmxBoneWeightBDEF4);
      3: (SDEF: TPmxBoneWeightSDEF);
      4: (QDEF: TPmxBoneWeightQDEF);
  end;

  TPmxVertex = record
    Position: PmxFloat3;
    Normal: PmxFloat3;
    UV: PmxFloat2;
    AdditionalUV: array of PmxFloat4;
    BoneWeightMethod: TPmxBoneWeightMethod; // 0:BDEF1 1:BDEF2 2:BDEF4 3:SDEF 4:QDEF(2.1g)
    BoneWeight: TPmxBoneWeight;
    MagnificationOfEdge: PmxFloat;
  end;
  PPmxVertex = ^TPmxVertex;

  TPmxMaterialRenderFlag = (rfTwoSide, rfGroundShadow, rfShadowMapping, rfSelfShadow,
    rfEdge, {2.1g->}rfVertexColor, rfPoint, rfLine);
  TPmxMaterialRenderFlags = set of TPmxMaterialRenderFlag;

  TPmxMaterialSphereMode = (smDisable, smMultiply, smAdd, smSubTexture);

  TPmxMaterialToonTexture = record
    case Integer of
      0: (ToonTextureIndex: PmxInt);
      1: (SharedToonTextureIndex: PmxByte);
  end;

  TPmxMaterial = record
    NameJp: string;
    NameEn: string;
    Diffuse: PmxFloat4;
    Specular: PmxFloat3;
    Shininess: PmxFloat;
    Ambient: PmxFloat3;
    RenderFlags: TPmxMaterialRenderFlags;
    EdgeColor: PmxFloat4;
    EdgeSize: PmxFloat;
    TextureIndex: PmxInt;
    SphereTextureIndex: PmxInt;
    SphereMode: TPmxMaterialSphereMode; // 0: 1:Z(sph) 2:Z(spa) 3:TueNX`(ǉUV1x,yUVQƂĒʏeNX``s)
    SharedToonFlag: PmxByte;
    ToonTexture: TPmxMaterialToonTexture;
    Memo: string;
    NumberOfVertex: PmxInt;
  end;
  PPmxMaterial = ^TPmxMaterial;

  TPmxBoneFlag = (bfConnection, bfCanRotate, bfCanTranslate, bfDisplay,
    bfCanOperate, bfIK, bfUnused{0x40}, bfAddLocal, bfAddRotation, bfAddTranslation,
    bfFixedAxis, bfLocalAxis, bfTransformAfterPhysics, bfTransformByExternalParent);
  TPmxBoneFlags = set of TPmxBoneFlag;

  TEulerType = (etXYZ, etYZX, etZXY);
  TAxisType = (atNone, atFixed, atFlowedX, atFlowedY, atFlowedZ);
  TPmxIKLink = record
    LinkBoneIndex: PmxInt;
    AngleLimit: PmxByte;
    LowerLimit: PmxFloat3;
    UpperLimit: PmxFloat3;
    Euler: TEulerType;
    AxisType: TAxisType;
  end;

  TPmxBone = record
    NameJp: string;
    NameEn: string;
    Position: PmxFloat3;
    ParentIndex: PmxInt;
    TransformationLevel: PmxInt;
    Flags: TPmxBoneFlags;
    PositionOffset: PmxFloat3;
    ConnectionBoneIndex: PmxInt;
    AddParentBoneIndex: PmxInt;
    AddPower: PmxFloat;
    FixedAxisDirection: PmxFloat3;
    LocalAxisXDirection: PmxFloat3;
    LocalAxisZDirection: PmxFloat3;
    ExternalParentKey: PmxInt;
    IKTargetBoneIndex: PmxInt;
    IKLoopCount: PmxInt;
    IKLoopLimitAngle: PmxFloat;
    IKLinks: array of TPmxIKLink;
  end;
  PPmxBone = ^TPmxBone;

  TPmxMorphTarget = (mtReserved, mtBrows, mtEyes, mtMouth, mtOther);
  TPmxMorphKind = (mkGroup, mkVertex, mkBone, mkUV, mkAddUV1, mkAddUV2,
    mkAddUV3, mkAddUV4, mkMaterial, {2.1g->}mkFlip, mkImpulse);

  TPmxMorphGroup = record
    MorphIndex: PmxInt;
    MorphRate: PmxFloat;
  end;

  TPmxMorphVertex = record
    VertexIndex: PmxInt;
    PositionOffset: PmxFloat3;
  end;

  TPmxMorphBone = record
    BoneIndex: PmxInt;
    TranslateValue: PmxFloat3;
    RotateValue: PmxFloat4;
  end;

  TPmxMorphUV = record
    VertexIndex: PmxInt;
    UVOffset: PmxFloat4;
  end;

  TPmxMorphMaterial = record
    MaterialIndex: PmxInt;
    OffsetType: PmxByte;
    Diffuse: PmxFloat4;
    Specular: PmxFloat3;
    Shininess: PmxFloat;
    Ambient: PmxFloat3;
    EdgeColor: PmxFloat4;
    EdgeSize: PmxFloat;
    TextureFactor: PmxFloat4;
    SphereTextureFactor: PmxFloat4;
    ToonTextureFactor: PmxFloat4;
  end;

  TPmxMorphFlip = TPmxMorphGroup;

  TPmxMorphImpulse = record
    RigidBodyIndex: PmxInt;
    LocalFlag: PmxByte;
    TranslationSpeed: PmxFloat3;
    RotationTorque: PmxFloat3;
  end;

  TPmxMorphData = record
    case TPmxMorphKind of
      mkGroup: (Group: TPmxMorphGroup);
      mkVertex: (Vertex: TPmxMorphVertex);
      mkBone: (Bone: TPmxMorphBone);
      mkUV, mkAddUV1, mkAddUV2, mkAddUV3, mkAddUV4: (UV: TPmxMorphUV);
      mkMaterial: (Material: TPmxMorphMaterial);
      mkFlip: (Flip: TPmxMorphFlip);
      mkImpulse: (Impulse: TPmxMorphImpulse);
  end;

  TPmxMorph = record
    NameJp: string;
    NameEn: string;
    Target: TPmxMorphTarget;
    Kind: TPmxMorphKind;
    Data: array of TPmxMorphData;
  end;
  PPmxMorph = ^TPmxMorph;

  TPmxDisplayFrameElement = record
    case Target: PmxByte of
      0: (BoneIndex: PmxInt);
      1: (MorphIndex: PmxInt);
  end;

  TPmxDisplayFrame = record
    NameJp: string;
    NameEn: string;
    SpecialFlag: PmxByte;
    Elements: array of TPmxDisplayFrameElement;
  end;
  PPmxDisplayFrame = ^TPmxDisplayFrame;

  TPmxRigidBodyShapeType = (stSphere, stBox, stCapsule);

  TPmxRigidBodyPhysicsOperation = (poStatic, poDynamic, poBoth);

  TPmxRigidBody = record
    NameJp: string;
    NameEn: string;
    RelatedBoneIndex: PmxInt;
    Group: PmxByte;
    CollisionlessGroupFlag: PmxUShort;
    Shape: TPmxRigidBodyShapeType; // 0: 1: 2:JvZ
    Size: PmxFloat3;
    Position: PmxFloat3;
    Rotation: PmxFloat3;
    Mass: PmxFloat;
    TranslateDecay: PmxFloat;
    RotateDecay: PmxFloat;
    Repulsion: PmxFloat;
    Friction: PmxFloat;
    PhysicsOperation: TPmxRigidBodyPhysicsOperation; // 0:{[Ǐ](static) 1:Z(dynamic) 2:Z + Boneʒu킹
  end;
  PPmxRigidBody = ^TPmxRigidBody;

  TPmxJointKind = (jkSpring6DOF, jk6DOF, jkP2P, jkConeTwist, jkSlider, jkHinge);

  TPmxJointSpring6DOF = record
    RigidBodyAIndex: PmxInt;
    RigidBodyBIndex: PmxInt;
    Position: PmxFloat3;
    Rotation: PmxFloat3;
    TranslateLowerLimit: PmxFloat3;
    TranslateUpperLimit: PmxFloat3;
    RotateLowerLimit: PmxFloat3;
    RotateUpperLimit: PmxFloat3;
    TranslationSpringConstant: PmxFloat3;
    RotationSpringConstant: PmxFloat3;
  end;

  TPmxJointData = record
    case TPmxJointKind of
      jkSpring6DOF: (Spring6DOF: TPmxJointSpring6DOF);
  end;

  TPmxJoint = record
    NameJp: string;
    NameEn: string;
    JointKind: TPmxJointKind;
    Data: TPmxJointData;
  end;
  PPmxJoint = ^TPmxJoint;

  TPmxSoftBodyFlag = (sfBLink, sfCluster, sfCrossLink);
  TPmxSoftBodyFlags = set of TPmxSoftBodyFlag;

  TPmxSoftBodyConfig = packed record
    VCF: PmxFloat;
    DP: PmxFloat;
    DG: PmxFloat;
    LF: PmxFloat;
    PR: PmxFloat;
    VC: PmxFloat;
    DF: PmxFloat;
    MT: PmxFloat;
    CHR: PmxFloat;
    KHR: PmxFloat;
    &SHR: PmxFloat;
    AHR: PmxFloat;
  end;

  TPmxSoftBodyCluster = packed record
    SRHR_CL: PmxFloat;
    SKHR_CL: PmxFloat;
    SSHR_CL: PmxFloat;
    SR_SPLT_CL: PmxFloat;
    SK_SPLT_CL: PmxFloat;
    SS_SPLT_CL: PmxFloat;
  end;

  TPmxSoftBodyIteration = packed record
    V_IT: PmxInt;
    P_IT: PmxInt;
    D_IT: PmxInt;
    C_IT: PmxInt;
  end;

  TPmxSoftBodyMaterial = packed record
    LST: PmxFloat;
    AST: PmxFloat;
    VST: PmxFloat;
  end;

  TPmxAnchorRigidBody = record
    RelatedRigidBodyIndex: PmxInt;
    RelatedVertexIndex: PmxInt;
    NearMode: PmxByte;
  end;

  TPmxPinVertex = record
    RelatedVertexIndex: PmxInt;
  end;

  TPmxSoftBody = record
    NameJp: string;
    NameEn: string;
    Shape: PmxByte; // 0:TriMesh 1:Rope
    RelatedMaterialIndex: PmxInt;
    Group: PmxByte;
    CollisionlessGroupFlag: PmxUShort;
    Flags: TPmxSoftBodyFlags;
    BLinkDistance: PmxInt;
    NumberOfCluster: PmxInt;
    TotalMass: PmxFloat;
    CollisionMargin: PmxFloat;
    AeroModel: PmxInt; // 0:V_Point, 1:V_TwoSided, 2:V_OneSided, 3:F_TwoSided, 4:F_OneSided
    Config: TPmxSoftBodyConfig;
    Cluster: TPmxSoftBodyCluster;
    Iteration: TPmxSoftBodyIteration;
    Material: TPmxSoftBodyMaterial;
    AnchorRigidBodies: array of TPmxAnchorRigidBody;
    PinVertices: array of TPmxPinVertex;
  end;
  PPmxSoftBody = ^TPmxSoftBody;

  TPmx = class
  private
    FModelNameJp: string;
    FModelNameEn: string;
    FCommentJp: string;
    FCommentEn: string;
    FVertexList: array of TPmxVertex;
    FFaceList: array of PmxInt;
    FTextureList: array of string;
    FMaterialList: array of TPmxMaterial;
    FBoneList: array of TPmxBone;
    FMorphList: array of TPmxMorph;
    FDisplayFrameList: array of TPmxDisplayFrame;
    FRigidBodyList: array of TPmxRigidBody;
    FJointList: array of TPmxJoint;
    FSoftBodyList: array of TPmxSoftBody;
    function GetVertexCount: Integer;
    function GetVertex(Index: Integer): PPmxVertex;
    function GetFaceCount: Integer;
    function GetFace(Index: Integer): PPmxInt3;
    function GetTextureCount: Integer;
    function GetTexture(Index: Integer): string;
    function GetMaterialCount: Integer;
    function GetMaterial(Index: Integer): PPmxMaterial;
    function GetBoneCount: Integer;
    function GetBone(Index: Integer): PPmxBone;
    function GetMorphCount: Integer;
    function GetMorph(Index: Integer): PPmxMorph;
    function GetDisplayFrameCount: Integer;
    function GetDisplayFrame(Index: Integer): PPmxDisplayFrame;
    function GetRigidBodyCount: Integer;
    function GetRigidBody(Index: Integer): PPmxRigidBody;
    function GetJointCount: Integer;
    function GetJoint(Index: Integer): PPmxJoint;
    function GetSoftBodyCount: Integer;
    function GetSoftBody(Index: Integer): PPmxSoftBody;
  public
    procedure LoadFromStream(Stream: TStream;
      RightHandedSystem: Boolean = False; Scale: Single = 1.0);
    procedure LoadFromFile(const FileName: string;
      RightHandedSystem: Boolean = False; Scale: Single = 1.0);

    function IndexOfBone(const AName: string): Integer;
    function IndexOfMorph(const AName: string): Integer;

    property ModelNameJp: string read FModelNameJp write FModelNameJp;
    property ModelNameEn: string read FModelNameEn write FModelNameEn;
    property CommentJp: string read FCommentJp write FCommentJp;
    property CommentEn: string read FCommentEn write FCommentEn;

    property VertexCount: Integer read GetVertexCount;
    property Vertices[Index: Integer]: PPmxVertex read GetVertex;
    property FaceCount: Integer read GetFaceCount;
    property Faces[Index: Integer]: PPmxInt3 read GetFace;
    property TextureCount: Integer read GetTextureCount;
    property Textures[Index: Integer]: string read GetTexture;
    property MaterialCount: Integer read GetMaterialCount;
    property Materials[Index: Integer]: PPmxMaterial read GetMaterial;
    property BoneCount: Integer read GetBoneCount;
    property Bones[Index: Integer]: PPmxBone read GetBone;
    property MorphCount: Integer read GetMorphCount;
    property Morphs[Index: Integer]: PPmxMorph read GetMorph;
    property DisplayFrameCount: Integer read GetDisplayFrameCount;
    property DisplayFrames[Index: Integer]: PPmxDisplayFrame read GetDisplayFrame;
    property RigidBodyCount: Integer read GetRigidBodyCount;
    property RigidBodies[Index: Integer]: PPmxRigidBody read GetRigidBody;
    property JointCount: Integer read GetJointCount;
    property Joints[Index: Integer]: PPmxJoint read GetJoint;
    property SoftBodyCount: Integer read GetSoftBodyCount;
    property SoftBodies[Index: Integer]: PPmxSoftBody read GetSoftBody;
  end;

  TVmdMotion = packed record // 111 Bytes // [V
    BoneName: array[0..14] of AnsiChar; // {[
    FrameNum: Longword; // t[ԍ(Ǎ݂͌̃t[ʒu0ƂΈʒu)
    Position: array[0..2] of Single; // ʒu
    Rotation: array[0..3] of Single; // Quaternion // ]
    Interpolation: array[0..63] of Byte; // [4][4][4] // ⊮
  end;
  PVmdMotion = ^TVmdMotion;

  TVmdSkinMotion = packed record // 23 Bytes // \
    MorphName: array[0..14] of AnsiChar; // \
    FrameNum: Longword; // t[ԍ
    Weight: Single; // \̐ݒl(\XC_[̒l)
  end;
  PVmdSkinMotion = ^TVmdSkinMotion;

  TVmd = class
  private
    FFrameCount: Longword;
    FMotionList: array of TVmdMotion;
    FSkinMotionList: array of TVmdSkinMotion;
    function GetMotionCount: Integer;
    function GetMotion(Index: Integer): PVmdMotion;
    function GetSkinMotionCount: Integer;
    function GetSkinMotion(Index: Integer): PVmdSkinMotion;
  public
    procedure LoadFromStream(Stream: TStream;
      RightHandedSystem: Boolean = False; Scale: Single = 1.0);
    procedure LoadFromFile(const FileName: string;
      RightHandedSystem: Boolean = False; Scale: Single = 1.0);

    property FrameCount: Longword read FFrameCount;
    property MotionCount: Integer read GetMotionCount;
    property Motions[Index: Integer]: PVmdMotion read GetMotion;
    property SkinMotionCount: Integer read GetSkinMotionCount;
    property SkinMotions[Index: Integer]: PVmdSkinMotion read GetSkinMotion;
  end;

const
  PMX_HEADER_ENCODING = 0; // GR[h | 0:UTF16 1:UTF8
  PMX_HEADER_ADDUV    = 1; // ǉUV | 0`4 ڍׂ͒_Q
  PMX_HEADER_VERTEX   = 2; // _IndexTCY | 1,2,4 ̂ꂩ
  PMX_HEADER_TEXTURE  = 3; // eNX`IndexTCY | 1,2,4 ̂ꂩ
  PMX_HEADER_MATERIAL = 4; // ގIndexTCY | 1,2,4 ̂ꂩ
  PMX_HEADER_BONE     = 5; // {[IndexTCY | 1,2,4 ̂ꂩ
  PMX_HEADER_MORPH    = 6; // [tIndexTCY | 1,2,4 ̂ꂩ
  PMX_HEADER_RIGID    = 7; // IndexTCY | 1,2,4 ̂ꂩ

  // GR[h
  PMX_ENCODING_UTF16 = 0;
  PMX_ENCODING_UTF8  = 1;

implementation

uses
  System.Math;

type
  TStreamHelper = class helper for TStream
  public
    function ReadVarInt(Size: Integer): Integer;
    function ReadUVarInt(Size: Integer): Integer;
    function ReadPMXString(Encoding: Integer): string;
  end;

function TStreamHelper.ReadVarInt(Size: Integer): Integer;
var
  b: ShortInt;
  w: SmallInt;
begin
  case Size of
    1:
    begin
      ReadBuffer(b, 1);
      Result := b;
    end;
    2:
    begin
      ReadBuffer(w, 2);
      Result := w;
    end;
    4:
    begin
      ReadBuffer(Result, 4);
    end;
  else
    raise Exception.Create('Ή̃TCY');
  end;
end;

function TStreamHelper.ReadUVarInt(Size: Integer): Integer;
var
  b: Byte;
  w: Word;
begin
  case Size of
    1:
    begin
      ReadBuffer(b, 1);
      Result := b;
    end;
    2:
    begin
      ReadBuffer(w, 2);
      Result := w;
    end;
    4:
    begin
      ReadBuffer(Result, 4);
    end;
  else
    raise Exception.Create('Ή̃TCY');
  end;
end;

function TStreamHelper.ReadPMXString(Encoding: Integer): string;
var
  len: Int32;
  bytes: TBytes;
begin
  ReadBuffer(len, SizeOf(len));
  SetLength(bytes, len);
  ReadBuffer(Pointer(bytes)^, len);

  if Encoding = PMX_ENCODING_UTF16 then
    Result := TEncoding.Unicode.GetString(bytes)
  else if Encoding = PMX_ENCODING_UTF8 then
    Result := TEncoding.UTF8.GetString(bytes)
  else
    raise Exception.CreateFmt('m̃GR[fBO (%d)', [Encoding]);
end;

{ TPmx }

procedure TPmx.LoadFromStream(Stream: TStream;
  RightHandedSystem: Boolean = False; Scale: Single = 1.0);
const
  PMX_SIG: array[0..3] of AnsiChar = 'PMX ';
var
  sig: array[0..3] of AnsiChar;
  version: PmxFloat;
  optionSize: PmxByte;
  headerOptions: array[0..7] of PmxByte;
  tmp: PmxInt;
  cnt, cntIK, cntMorphData, cntElement, cntRigid, cntPin: PmxInt;
  i, iIK, iMorphData, iElement, iRigid, iPin: NativeInt;
  lower, upper: PmxFloat3;

  procedure AdjustVector(var V: PmxFloat3);
  begin
    if Scale <> 1.0 then
    begin
      V.X := V.X * Scale;
      V.Y := V.Y * Scale;
      V.Z := V.Z * Scale;
    end;
    if RightHandedSystem then
    begin
      V.Z := -V.Z;
    end;
  end;

begin
  // PMXwb_
  Stream.ReadBuffer(sig, SizeOf(sig));
  if sig <> PMX_SIG then
    raise Exception.CreateFmt('m̃VOl` (%s)', [sig]);
  Stream.ReadBuffer(version, SizeOf(version));
  if (version <> 2.0) and (version <> 2.1) then
    raise Exception.CreateFmt('m̃o[W (%f)', [version]);

  Stream.ReadBuffer(optionSize, SizeOf(optionSize));
  if optionSize <> 8 then
    raise Exception.Create('8ȊÕwb_oCgTCYɂ͖Ή');

  Stream.ReadBuffer(headerOptions, optionSize);

  // f
  FModelNameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
  FModelNameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
  FCommentJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
  FCommentEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);

  // _
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FVertexList, cnt);
  for i := 0 to cnt-1 do
  begin
    Stream.ReadBuffer(FVertexList[i].Position, SizeOf(PmxFloat3));
    Stream.ReadBuffer(FVertexList[i].Normal, SizeOf(PmxFloat3));
    Stream.ReadBuffer(FVertexList[i].UV, SizeOf(PmxFloat2));
    SetLength(FVertexList[i].AdditionalUV, headerOptions[PMX_HEADER_ADDUV]);
    Stream.ReadBuffer(Pointer(FVertexList[i].AdditionalUV)^, headerOptions[PMX_HEADER_ADDUV] * SizeOf(PmxFloat4));
    Stream.ReadBuffer(FVertexList[i].BoneWeightMethod, SizeOf(TPmxBoneWeightMethod));
    case FVertexList[i].BoneWeightMethod of
      wmBDEF1:
      begin
        FVertexList[i].BoneWeight.BDEF1.BoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
      end;
      wmBDEF2:
      begin
        with FVertexList[i].BoneWeight.BDEF2 do
        begin
          Bone1Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Bone2Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Stream.ReadBuffer(Bone1Weight, SizeOf(PmxFloat));
        end;
      end;
      wmBDEF4,
      wmQDEF:
      begin
        with FVertexList[i].BoneWeight.BDEF4 do
        begin
          Bone1Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Bone2Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Bone3Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Bone4Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Stream.ReadBuffer(Bone1Weight, SizeOf(PmxFloat));
          Stream.ReadBuffer(Bone2Weight, SizeOf(PmxFloat));
          Stream.ReadBuffer(Bone3Weight, SizeOf(PmxFloat));
          Stream.ReadBuffer(Bone4Weight, SizeOf(PmxFloat));
        end;
      end;
      wmSDEF:
      begin
        with FVertexList[i].BoneWeight.SDEF do
        begin
          Bone1Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Bone2Index := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
          Stream.ReadBuffer(Bone1Weight, SizeOf(PmxFloat));
          Stream.ReadBuffer(SDEF_C, SizeOf(PmxFloat3));
          Stream.ReadBuffer(SDEF_R0, SizeOf(PmxFloat3));
          Stream.ReadBuffer(SDEF_R1, SizeOf(PmxFloat3));
        end;
      end;
    end;
    Stream.ReadBuffer(FVertexList[i].MagnificationOfEdge, SizeOf(PmxFloat));

    // 
    AdjustVector(FVertexList[i].Position);
    if RightHandedSystem then
    begin
      FVertexList[i].Normal.Z := -FVertexList[i].Normal.Z;
    end;
  end;

  // 
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FFaceList, cnt);
  for i := 0 to cnt-1 do
  begin
    FFaceList[i] := Stream.ReadUVarInt(headerOptions[PMX_HEADER_VERTEX]);
  end;
  if RightHandedSystem then
  begin
    for i := 0 to cnt div 3 - 1 do
    begin
      tmp := FFaceList[i * 3 + 1];
      FFaceList[i * 3 + 1] := FFaceList[i * 3 + 2];
      FFaceList[i * 3 + 2] := tmp;
    end;
  end;

  // eNX`
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FTextureList, cnt);
  for i := 0 to cnt-1 do
  begin
    FTextureList[i] := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FTextureList[i] := FTextureList[i].Replace('/', '\'); // TODO: WindowsȊOl
  end;

  // ގ
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FMaterialList, cnt);
  for i := 0 to cnt-1 do
  begin
    FMaterialList[i].NameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FMaterialList[i].NameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    Stream.ReadBuffer(FMaterialList[i].Diffuse, SizeOf(PmxFloat4));
    Stream.ReadBuffer(FMaterialList[i].Specular, SizeOf(PmxFloat3));
    Stream.ReadBuffer(FMaterialList[i].Shininess, SizeOf(PmxFloat));
    Stream.ReadBuffer(FMaterialList[i].Ambient, SizeOf(PmxFloat3));
    Stream.ReadBuffer(FMaterialList[i].RenderFlags, SizeOf(TPmxMaterialRenderFlags));
    Stream.ReadBuffer(FMaterialList[i].EdgeColor, SizeOf(PmxFloat4));
    Stream.ReadBuffer(FMaterialList[i].EdgeSize, SizeOf(PmxFloat));
    FMaterialList[i].TextureIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_TEXTURE]);
    FMaterialList[i].SphereTextureIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_TEXTURE]);
    Stream.ReadBuffer(FMaterialList[i].SphereMode, SizeOf(TPmxMaterialSphereMode));
    Stream.ReadBuffer(FMaterialList[i].SharedToonFlag, SizeOf(PmxByte));
    if FMaterialList[i].SharedToonFlag = 0 then
      FMaterialList[i].ToonTexture.ToonTextureIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_TEXTURE])
    else
      Stream.ReadBuffer(FMaterialList[i].ToonTexture.SharedToonTextureIndex, SizeOf(PmxByte));
    FMaterialList[i].Memo := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    Stream.ReadBuffer(FMaterialList[i].NumberOfVertex, SizeOf(PmxInt));
  end;

  // {[
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FBoneList, cnt);
  for i := 0 to cnt-1 do
  begin
    FBoneList[i].NameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FBoneList[i].NameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    Stream.ReadBuffer(FBoneList[i].Position, SizeOf(PmxFloat3));
    FBoneList[i].ParentIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
    Stream.ReadBuffer(FBoneList[i].TransformationLevel, SizeOf(PmxInt));
    Stream.ReadBuffer(FBoneList[i].Flags, SizeOf(TPmxBoneFlags));
    if not (bfConnection in FBoneList[i].Flags) then
      Stream.ReadBuffer(FBoneList[i].PositionOffset, SizeOf(PmxFloat3))
    else
      FBoneList[i].ConnectionBoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
    if FBoneList[i].Flags * [bfAddRotation, bfAddTranslation] <> [] then
    begin
      FBoneList[i].AddParentBoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
      Stream.ReadBuffer(FBoneList[i].AddPower, SizeOf(PmxFloat));
    end;
    if bfFixedAxis in FBoneList[i].Flags then
      Stream.ReadBuffer(FBoneList[i].FixedAxisDirection, SizeOf(PmxFloat3));
    if bfLocalAxis in FBoneList[i].Flags then
    begin
      Stream.ReadBuffer(FBoneList[i].LocalAxisXDirection, SizeOf(PmxFloat3));
      Stream.ReadBuffer(FBoneList[i].LocalAxisZDirection, SizeOf(PmxFloat3));
    end;
    if bfTransformByExternalParent in FBoneList[i].Flags then
      Stream.ReadBuffer(FBoneList[i].ExternalParentKey, SizeOf(PmxInt));
    if bfIK in FBoneList[i].Flags then
    begin
      FBoneList[i].IKTargetBoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
      Stream.ReadBuffer(FBoneList[i].IKLoopCount, SizeOf(PmxInt));
      Stream.ReadBuffer(FBoneList[i].IKLoopLimitAngle, SizeOf(PmxFloat));
      Stream.ReadBuffer(cntIK, SizeOf(PmxInt));
      SetLength(FBoneList[i].IKLinks, cntIK);
      for iIK := 0 to cntIK-1 do
      begin
        FBoneList[i].IKLinks[iIK].LinkBoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
        Stream.ReadBuffer(FBoneList[i].IKLinks[iIK].AngleLimit, SizeOf(PmxByte));
        if FBoneList[i].IKLinks[iIK].AngleLimit = 1 then
        begin
          Stream.ReadBuffer(lower, SizeOf(PmxFloat3));
          Stream.ReadBuffer(upper, SizeOf(PmxFloat3));

          // K
          with FBoneList[i].IKLinks[iIK] do
          begin
            if RightHandedSystem then
            begin
              lower.X := -lower.X;
              lower.Y := -lower.Y;
              upper.X := -upper.X;
              upper.Y := -upper.Y;
            end;

            LowerLimit.X := Min(lower.X, upper.X);
            LowerLimit.Y := Min(lower.Y, upper.Y);
            LowerLimit.Z := Min(lower.Z, upper.Z);
            UpperLimit.X := Max(lower.X, upper.X);
            UpperLimit.Y := Max(lower.Y, upper.Y);
            UpperLimit.Z := Max(lower.Z, upper.Z);

            if (LowerLimit.X > -Pi/2) and (UpperLimit.X < Pi/2) then
              Euler := etZXY
            else if (LowerLimit.Y > -Pi/2) and (UpperLimit.Y < Pi/2) then
              Euler := etXYZ
            else
              Euler := etYZX;

            AxisType := atNone;
            if (LowerLimit.X = 0) and (UpperLimit.X = 0) and (LowerLimit.Y = 0) and (UpperLimit.Y = 0) and (LowerLimit.Z = 0) and (UpperLimit.Z = 0) then
              AxisType := atFixed
            else if (LowerLimit.Y = 0) and (UpperLimit.Y = 0) and (LowerLimit.Z = 0) and (UpperLimit.Z = 0) then
              AxisType := atFlowedX
            else if (LowerLimit.X = 0) and (UpperLimit.X = 0) and (LowerLimit.Z = 0) and (UpperLimit.Z = 0) then
              AxisType := atFlowedY
            else if (LowerLimit.X = 0) and (UpperLimit.X = 0) and (LowerLimit.Y = 0) and (UpperLimit.Y = 0) then
              AxisType := atFlowedZ;
          end;
        end;
      end;
    end;

    // 
    AdjustVector(FBoneList[i].Position);
    if not (bfConnection in FBoneList[i].Flags) then
      AdjustVector(FBoneList[i].PositionOffset);
  end;

  // [t
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FMorphList, cnt);
  for i := 0 to cnt-1 do
  begin
    FMorphList[i].NameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FMorphList[i].NameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    Stream.ReadBuffer(FMorphList[i].Target, SizeOf(TPmxMorphTarget));
    Stream.ReadBuffer(FMorphList[i].Kind, SizeOf(TPmxMorphKind));
    Stream.ReadBuffer(cntMorphData, SizeOf(PmxInt));
    SetLength(FMorphList[i].Data, cntMorphData);
    case FMorphList[i].Kind of
      mkGroup,
      mkFlip:
        for iMorphData := 0 to cntMorphData-1 do
        begin
          with FMorphList[i].Data[iMorphData].Group do
          begin
            MorphIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_MORPH]);
            Stream.ReadBuffer(MorphRate, SizeOf(PmxFloat));
          end;
        end;
      mkVertex:
        for iMorphData := 0 to cntMorphData-1 do
        begin
          with FMorphList[i].Data[iMorphData].Vertex do
          begin
            VertexIndex := Stream.ReadUVarInt(headerOptions[PMX_HEADER_VERTEX]);
            Stream.ReadBuffer(PositionOffset, SizeOf(PmxFloat3));

            // 
            AdjustVector(PositionOffset);
          end;
        end;
      mkBone:
        for iMorphData := 0 to cntMorphData-1 do
        begin
          with FMorphList[i].Data[iMorphData].Bone do
          begin
            BoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
            Stream.ReadBuffer(TranslateValue, SizeOf(PmxFloat3));
            Stream.ReadBuffer(RotateValue, SizeOf(PmxFloat4));

            // 
            AdjustVector(TranslateValue);
            if RightHandedSystem then
            begin
              RotateValue.Z := -RotateValue.Z;
              RotateValue.W := -RotateValue.W;
            end;
          end;
        end;
      mkUV, mkAddUV1, mkAddUV2, mkAddUV3, mkAddUV4:
        for iMorphData := 0 to cntMorphData-1 do
        begin
          with FMorphList[i].Data[iMorphData].UV do
          begin
            VertexIndex := Stream.ReadUVarInt(headerOptions[PMX_HEADER_VERTEX]);
            Stream.ReadBuffer(UVOffset, SizeOf(PmxFloat4));
          end;
        end;
      mkMaterial:
        for iMorphData := 0 to cntMorphData-1 do
        begin
          with FMorphList[i].Data[iMorphData].Material do
          begin
            MaterialIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_MATERIAL]);
            Stream.ReadBuffer(OffsetType, SizeOf(PmxByte));
            Stream.ReadBuffer(Diffuse, SizeOf(PmxFloat4));
            Stream.ReadBuffer(Specular, SizeOf(PmxFloat3));
            Stream.ReadBuffer(Shininess, SizeOf(PmxFloat));
            Stream.ReadBuffer(Ambient, SizeOf(PmxFloat3));
            Stream.ReadBuffer(EdgeColor, SizeOf(PmxFloat4));
            Stream.ReadBuffer(EdgeSize, SizeOf(PmxFloat));
            Stream.ReadBuffer(TextureFactor, SizeOf(PmxFloat4));
            Stream.ReadBuffer(SphereTextureFactor, SizeOf(PmxFloat4));
            Stream.ReadBuffer(ToonTextureFactor, SizeOf(PmxFloat4));
          end;
        end;
      mkImpulse:
        for iMorphData := 0 to cntMorphData-1 do
        begin
          with FMorphList[i].Data[iMorphData].Impulse do
          begin
            RigidBodyIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_RIGID]);
            Stream.ReadBuffer(LocalFlag, SizeOf(PmxByte));
            Stream.ReadBuffer(TranslationSpeed, SizeOf(PmxFloat3));
            Stream.ReadBuffer(RotationTorque, SizeOf(PmxFloat3));
          end;
        end;
    end;
  end;

  // \g
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FDisplayFrameList, cnt);
  for i := 0 to cnt-1 do
  begin
    FDisplayFrameList[i].NameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FDisplayFrameList[i].NameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    Stream.ReadBuffer(FDisplayFrameList[i].SpecialFlag, SizeOf(PmxByte));
    Stream.ReadBuffer(cntElement, SizeOf(PmxInt));
    SetLength(FDisplayFrameList[i].Elements, cntElement);
    for iElement := 0 to cntElement-1 do
    begin
      Stream.ReadBuffer(FDisplayFrameList[i].Elements[iElement].Target, SizeOf(PmxByte));
      case FDisplayFrameList[i].Elements[iElement].Target of
        0: FDisplayFrameList[i].Elements[iElement].BoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
        1: FDisplayFrameList[i].Elements[iElement].MorphIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_MORPH]);
      end;
    end;
  end;

  // 
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FRigidBodyList, cnt);
  for i := 0 to cnt-1 do
  begin
    FRigidBodyList[i].NameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FRigidBodyList[i].NameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FRigidBodyList[i].RelatedBoneIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_BONE]);
    Stream.ReadBuffer(FRigidBodyList[i].Group, SizeOf(PmxByte));
    Stream.ReadBuffer(FRigidBodyList[i].CollisionlessGroupFlag, SizeOf(PmxUShort));
    Stream.ReadBuffer(FRigidBodyList[i].Shape, SizeOf(TPmxRigidBodyShapeType));
    Stream.ReadBuffer(FRigidBodyList[i].Size, SizeOf(PmxFloat3));
    Stream.ReadBuffer(FRigidBodyList[i].Position, SizeOf(PmxFloat3));
    Stream.ReadBuffer(FRigidBodyList[i].Rotation, SizeOf(PmxFloat3));
    Stream.ReadBuffer(FRigidBodyList[i].Mass, SizeOf(PmxFloat));
    Stream.ReadBuffer(FRigidBodyList[i].TranslateDecay, SizeOf(PmxFloat));
    Stream.ReadBuffer(FRigidBodyList[i].RotateDecay, SizeOf(PmxFloat));
    Stream.ReadBuffer(FRigidBodyList[i].Repulsion, SizeOf(PmxFloat));
    Stream.ReadBuffer(FRigidBodyList[i].Friction, SizeOf(PmxFloat));
    Stream.ReadBuffer(FRigidBodyList[i].PhysicsOperation, SizeOf(TPmxRigidBodyPhysicsOperation));

    // 
    AdjustVector(FRigidBodyList[i].Position);
    if RightHandedSystem then
    begin
      FRigidBodyList[i].Rotation.Z := -FRigidBodyList[i].Rotation.Z;
    end;
  end;

  // Joint
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FJointList, cnt);
  for i := 0 to cnt-1 do
  begin
    FJointList[i].NameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FJointList[i].NameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    Stream.ReadBuffer(FJointList[i].JointKind, SizeOf(TPmxJointKind));
    case FJointList[i].JointKind of
      jkSpring6DOF,
      jk6DOF,
      jkP2P,
      jkConeTwist,
      jkSlider,
      jkHinge:
      begin
        with FJointList[i].Data.Spring6DOF do
        begin
          RigidBodyAIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_RIGID]);
          RigidBodyBIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_RIGID]);
          Stream.ReadBuffer(Position, SizeOf(PmxFloat3));
          Stream.ReadBuffer(Rotation, SizeOf(PmxFloat3));
          Stream.ReadBuffer(TranslateLowerLimit, SizeOf(PmxFloat3));
          Stream.ReadBuffer(TranslateUpperLimit, SizeOf(PmxFloat3));
          Stream.ReadBuffer(RotateLowerLimit, SizeOf(PmxFloat3));
          Stream.ReadBuffer(RotateUpperLimit, SizeOf(PmxFloat3));
          Stream.ReadBuffer(TranslationSpringConstant, SizeOf(PmxFloat3));
          Stream.ReadBuffer(RotationSpringConstant, SizeOf(PmxFloat3));

          // 
          AdjustVector(Position);
          AdjustVector(TranslateLowerLimit);
          AdjustVector(TranslateUpperLimit);
          AdjustVector(TranslationSpringConstant);
          if RightHandedSystem then
          begin
            Rotation.Z := -Rotation.Z;
            RotateLowerLimit.Z := -RotateLowerLimit.Z;
            RotateUpperLimit.Z := -RotateUpperLimit.Z;
            RotationSpringConstant.Z := -RotationSpringConstant.Z;
          end;
        end;
      end;
    end;
  end;

  if version <= 2.0 then Exit;

  // SoftBody
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FSoftBodyList, cnt);
  for i := 0 to cnt-1 do
  begin
    FSoftBodyList[i].NameJp := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    FSoftBodyList[i].NameEn := Stream.ReadPMXString(headerOptions[PMX_HEADER_ENCODING]);
    Stream.ReadBuffer(FSoftBodyList[i].Shape, SizeOf(PmxByte));
    FSoftBodyList[i].RelatedMaterialIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_MATERIAL]);
    Stream.ReadBuffer(FSoftBodyList[i].Group, SizeOf(PmxByte));
    Stream.ReadBuffer(FSoftBodyList[i].CollisionlessGroupFlag, SizeOf(PmxUShort));
    Stream.ReadBuffer(FSoftBodyList[i].Flags, SizeOf(TPmxSoftBodyFlags));
    Stream.ReadBuffer(FSoftBodyList[i].BLinkDistance, SizeOf(PmxInt));
    Stream.ReadBuffer(FSoftBodyList[i].NumberOfCluster, SizeOf(PmxInt));
    Stream.ReadBuffer(FSoftBodyList[i].TotalMass, SizeOf(PmxFloat));
    Stream.ReadBuffer(FSoftBodyList[i].CollisionMargin, SizeOf(PmxFloat));
    Stream.ReadBuffer(FSoftBodyList[i].AeroModel, SizeOf(PmxInt));
    Stream.ReadBuffer(FSoftBodyList[i].Config, SizeOf(TPmxSoftBodyConfig));
    Stream.ReadBuffer(FSoftBodyList[i].Cluster, SizeOf(TPmxSoftBodyCluster));
    Stream.ReadBuffer(FSoftBodyList[i].Iteration, SizeOf(TPmxSoftBodyIteration));
    Stream.ReadBuffer(FSoftBodyList[i].Material, SizeOf(TPmxSoftBodyMaterial));
    Stream.ReadBuffer(cntRigid, SizeOf(PmxInt));
    SetLength(FSoftBodyList[i].AnchorRigidBodies, cntRigid);
    for iRigid := 0 to cntRigid-1 do
    begin
      FSoftBodyList[i].AnchorRigidBodies[iRigid].RelatedRigidBodyIndex := Stream.ReadVarInt(headerOptions[PMX_HEADER_RIGID]);
      FSoftBodyList[i].AnchorRigidBodies[iRigid].RelatedVertexIndex := Stream.ReadUVarInt(headerOptions[PMX_HEADER_VERTEX]);
      Stream.ReadBuffer(FSoftBodyList[i].AnchorRigidBodies[iRigid].NearMode, SizeOf(PmxByte));
    end;
    Stream.ReadBuffer(cntPin, SizeOf(PmxInt));
    SetLength(FSoftBodyList[i].PinVertices, cntPin);
    for iPin := 0 to cntPin-1 do
    begin
      FSoftBodyList[i].PinVertices[iPin].RelatedVertexIndex := Stream.ReadUVarInt(headerOptions[PMX_HEADER_VERTEX]);
    end;
  end;
end;

procedure TPmx.LoadFromFile(const FileName: string;
  RightHandedSystem: Boolean = False; Scale: Single = 1.0);
var
  ms: TMemoryStream;
begin
  ms := TMemoryStream.Create;
  try
    ms.LoadFromFile(FileName);
    LoadFromStream(ms, RightHandedSystem, Scale);
  finally
    ms.Free;
  end;
end;

function TPmx.IndexOfBone(const AName: string): Integer;
begin
  for Result := Low(FBoneList) to High(FBoneList) do
    if FBoneList[Result].NameJp = AName then Exit;
  Result := -1;
end;

function TPmx.IndexOfMorph(const AName: string): Integer;
begin
  for Result := Low(FMorphList) to High(FMorphList) do
    if FMorphList[Result].NameJp = AName then Exit;
  Result := -1;
end;

function TPmx.GetVertexCount: Integer;
begin
  Result := Length(FVertexList);
end;

function TPmx.GetVertex(Index: Integer): PPmxVertex;
begin
  Result := @FVertexList[Index];
end;

function TPmx.GetFaceCount: Integer;
begin
  Result := Length(FFaceList) div 3;
end;

function TPmx.GetFace(Index: Integer): PPmxInt3;
begin
  Result := PPmxInt3(@FFaceList[Index*3]);
end;

function TPmx.GetTextureCount: Integer;
begin
  Result := Length(FTextureList);
end;

function TPmx.GetTexture(Index: Integer): string;
begin
  Result := FTextureList[Index];
end;

function TPmx.GetMaterialCount: Integer;
begin
  Result := Length(FMaterialList);
end;

function TPmx.GetMaterial(Index: Integer): PPmxMaterial;
begin
  Result := @FMaterialList[Index];
end;

function TPmx.GetBoneCount: Integer;
begin
  Result := Length(FBoneList);
end;

function TPmx.GetBone(Index: Integer): PPmxBone;
begin
  Result := @FBoneList[Index];
end;

function TPmx.GetMorphCount: Integer;
begin
  Result := Length(FMorphList);
end;

function TPmx.GetMorph(Index: Integer): PPmxMorph;
begin
  Result := @FMorphList[Index];
end;

function TPmx.GetDisplayFrameCount: Integer;
begin
  Result := Length(FDisplayFrameList);
end;

function TPmx.GetDisplayFrame(Index: Integer): PPmxDisplayFrame;
begin
  Result := @FDisplayFrameList[Index];
end;

function TPmx.GetRigidBodyCount: Integer;
begin
  Result := Length(FRigidBodyList);
end;

function TPmx.GetRigidBody(Index: Integer): PPmxRigidBody;
begin
  Result := @FRigidBodyList[Index];
end;

function TPmx.GetJointCount: Integer;
begin
  Result := Length(FJointList);
end;

function TPmx.GetJoint(Index: Integer): PPmxJoint;
begin
  Result := @FJointList[Index];
end;

function TPmx.GetSoftBodyCount: Integer;
begin
  Result := Length(FSoftBodyList);
end;

function TPmx.GetSoftBody(Index: Integer): PPmxSoftBody;
begin
  Result := @FSoftBodyList[Index];
end;

{ TVmd }

procedure TVmd.LoadFromStream(Stream: TStream;
  RightHandedSystem: Boolean = False; Scale: Single = 1.0);
var
  header: array[0..29] of AnsiChar;
  modelName: array[0..19] of AnsiChar;
  i, cnt: Integer;
begin
  Stream.ReadBuffer(header, SizeOf(header));
  if header <> 'Vocaloid Motion Data 0002' then
    raise Exception.Create('Ή̃wb_');
  Stream.ReadBuffer(modelName, SizeOf(modelName));

  // [Vf[^
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FMotionList, cnt);
  Stream.ReadBuffer(Pointer(FMotionList)^, SizeOf(TVmdMotion)*cnt);

  // \f[^
  Stream.ReadBuffer(cnt, SizeOf(cnt));
  SetLength(FSkinMotionList, cnt);
  Stream.ReadBuffer(Pointer(FSkinMotionList)^, SizeOf(TVmdSkinMotion)*cnt);

  // TODO: ̑f[^̓ǂݍ

  // ől
  FFrameCount := 0;
  for i := Low(FMotionList) to High(FMotionList) do
  begin
    if FMotionList[i].FrameNum+1 > FFrameCount then
      FFrameCount := FMotionList[i].FrameNum+1;
  end;
  for i := Low(FSkinMotionList) to High(FSkinMotionList) do
  begin
    if FSkinMotionList[i].FrameNum+1 > FFrameCount then
      FFrameCount := FSkinMotionList[i].FrameNum+1;
  end;

  // XP[ύX
  if Scale <> 1.0 then
  begin
    for i := Low(FMotionList) to High(FMotionList) do
    begin
      FMotionList[i].Position[0] := FMotionList[i].Position[0] * Scale;
      FMotionList[i].Position[1] := FMotionList[i].Position[1] * Scale;
      FMotionList[i].Position[2] := FMotionList[i].Position[2] * Scale;
    end;
  end;

  // EWnɕϊ
  if RightHandedSystem then
  begin
    for i := Low(FMotionList) to High(FMotionList) do
    begin
      FMotionList[i].Position[2] := -FMotionList[i].Position[2];

      FMotionList[i].Rotation[2] := -FMotionList[i].Rotation[2];
      FMotionList[i].Rotation[3] := -FMotionList[i].Rotation[3];
    end;
  end;
end;

procedure TVmd.LoadFromFile(const FileName: string;
  RightHandedSystem: Boolean = False; Scale: Single = 1.0);
var
  ms: TMemoryStream;
begin
  ms := TMemoryStream.Create;
  try
    ms.LoadFromFile(FileName);
    LoadFromStream(ms, RightHandedSystem, Scale);
  finally
    ms.Free;
  end;
end;

function TVmd.GetMotionCount: Integer;
begin
  Result := Length(FMotionList);
end;

function TVmd.GetMotion(Index: Integer): PVmdMotion;
begin
  Result := @FMotionList[Index];
end;

function TVmd.GetSkinMotionCount: Integer;
begin
  Result := Length(FSkinMotionList);
end;

function TVmd.GetSkinMotion(Index: Integer): PVmdSkinMotion;
begin
  Result := @FSkinMotionList[Index];
end;

end.
